package liber

import (
	"io"
	"io/ioutil"
	"os"
	"testing"
)

type testDatabase struct {
	db        *Database
	path      string
	refbookid BookId
}

func (td *testDatabase) Close() {
	td.db.Close()
	os.RemoveAll(td.path)
}

func newTestDatabase(t *testing.T) (*testDatabase, *Database) {
	path, _ := ioutil.TempDir("", "testdb-")
	db, err := NewDb(path)
	if err != nil {
		t.Fatalf("NewDb(): %v", err)
	}

	book := testEbook()
	if err = db.PutBook(book); err != nil {
		t.Fatalf("PutBook(): %v", err)
	}
	if err = db.PutFile(testEpubFile(path, book.Id)); err != nil {
		t.Fatalf("PutFile(): %v", err)
	}

	return &testDatabase{db: db, path: path, refbookid: book.Id}, db
}

func testEpubFile(dir string, bookid BookId) *File {
	f, _ := ioutil.TempFile(dir, "ebook-")
	io.WriteString(f, "epub\n")
	f.Close()
	return &File{
		Id:       bookid,
		Path:     f.Name(),
		FileType: ".epub",
		Size:     4,
	}
}

func testEbook() *Book {
	return &Book{
		Id: NewID(),
		Metadata: &Metadata{
			Title:       "20,000 Leagues under the sea",
			Creator:     []string{"Jules Verne"},
			ISBN:        []string{"1234"},
			Description: "A pretty cool book.",
		},
	}
}

func TestDatabase_Get(t *testing.T) {
	td, db := newTestDatabase(t)
	defer td.Close()

	_, err := db.GetBook(td.refbookid)
	if err != nil {
		t.Fatalf("GetBook(%d): %v", td.refbookid, err)
	}
	files, err := db.GetBookFiles(td.refbookid)
	if err != nil {
		t.Fatalf("GetBookFiles(%d): %v", td.refbookid, err)
	}
	if len(files) != 1 {
		t.Fatalf("GetBookFiles(%d) bad result: %v", td.refbookid, files)
	}
}

func TestDatabase_BookFileRelation(t *testing.T) {
	td, db := newTestDatabase(t)
	defer td.Close()

	checkFiles := func(tag string, n int) []*File {
		files, err := db.GetBookFiles(td.refbookid)
		if err != nil {
			t.Fatalf("GetBookFiles@%s(%d): %v", tag, td.refbookid, err)
		}
		if len(files) != n {
			t.Fatalf("GetBookFiles@%s(%d) bad result (exp. len=%d): %v", tag, td.refbookid, n, files)
		}
		return files
	}

	files := checkFiles("init", 1)
	file0 := files[0]
	file1 := testEpubFile(td.path, td.refbookid)
	db.PutFile(file1)
	checkFiles("post_add", 2)

	db.DeleteFile(file1.Path)
	checkFiles("post_delete", 1)

	db.DeleteFile(file0.Path)
	checkFiles("post_delete_2", 0)

	if _, err := db.GetBook(td.refbookid); err == nil {
		t.Fatal("Book was not removed when n.files==0")
	}
}

func TestDatabase_Search(t *testing.T) {
	td, db := newTestDatabase(t)
	defer td.Close()

	r, err := db.Search("jules verne", 0, 100)
	if err != nil {
		t.Fatal(err)
	}
	if r.NumResults != 1 {
		t.Fatalf("NumResults != 1 - %#v", r)
	}
	if len(r.Results) != 1 {
		t.Fatalf("len(Results) != 1 - %#v", r)
	}

	r, err = db.Search("italo calvino", 0, 100)
	if err != nil {
		t.Fatal(err)
	}
	if r.NumResults > 0 {
		t.Fatalf("unexpected results: %v", r)
	}
}

func TestDatabase_Find(t *testing.T) {
	td, db := newTestDatabase(t)
	defer td.Close()

	// Find using ISBN.
	book := testEbook()
	if _, err := db.Find(book.Metadata); err != nil {
		t.Fatal("With ISBN: ", err)
	}

	// Find using title/author.
	book.Metadata.ISBN = nil
	if _, err := db.Find(book.Metadata); err != nil {
		t.Fatal("With title/author: ", err)
	}

	// Find only using title (should fail).
	book.Metadata.Creator = nil
	if result, err := db.Find(book.Metadata); err == nil {
		t.Fatalf("Title only: no error, result = %v", result)
	}
}

func TestDatabase_Autocomplete(t *testing.T) {
	td, db := newTestDatabase(t)
	defer td.Close()

	r, err := db.Autocomplete("jul")
	if err != nil {
		t.Fatal(err)
	}
	if r.NumResults == 0 {
		t.Error("No results")
	}

	r, err = db.Autocomplete("foo")
	if err != nil {
		t.Fatal(err)
	}
	if r.NumResults != 0 {
		t.Errorf("Unexpected results: %v", r)
	}
}
